# -*- coding: utf-8 -*-
"""Implement geodisWS."""
import requests
from lxml import objectify
from jinja2 import Environment, PackageLoader
from roulier.transport import Transport
from roulier.ws_tools import remove_empty_tags, get_parts
from roulier.exception import CarrierError
import logging

log = logging.getLogger(__name__)


class GeodisTransport(Transport):
    """Implement Geodis WS communication."""

    GEODIS_WS = "http://espace.geodis.com/geolabel/services/ImpressionEtiquette"  # nopep8
    GEODIS_WS_TEST = "http://espace.recette.geodis.com/geolabel/services/ImpressionEtiquette"  # nopep8

    def send(self, payload):
        """Call this function.

        Args:
            payload.body: XML in a string
            payload.header : auth
        Return:
            {
                response: (Requests.response)
                body: XML response (without soap)
                parts: dict of attachments
            }
        """
        body = payload['body']
        headers = payload['headers']
        is_test = payload['is_test']
        soap_message = self.soap_wrap(body, headers)
        response = self.send_request(soap_message, is_test)
        log.info('WS response time %s' % response.elapsed.total_seconds())
        return self.handle_response(response)

    def soap_wrap(self, body, auth):
        """Wrap body in a soap:Enveloppe."""
        env = Environment(
            loader=PackageLoader('roulier', '/carriers/geodis/templates'),
            extensions=['jinja2.ext.with_'])

        template = env.get_template("geodis_soap.xml")
        body_stripped = remove_empty_tags(body)
        header_template = env.get_template("geodis_header.xml")
        header_xml = header_template.render(auth=auth)
        data = template.render(body=body_stripped, header=header_xml)
        return data.encode('utf8')

    def send_request(self, body, is_test):
        """Send body to geodis WS."""
        ws_url = self.GEODIS_WS_TEST if is_test else self.GEODIS_WS
        return requests.post(
            ws_url,
            headers={
                'content-type': 'text/xml',
                'SOAPAction': '<SOAP Action>'
            },
            data=body)

    def handle_500(self, response):
        """Handle reponse in case of ERROR 500 type."""
        # TODO : put a try catch (like wrong server)
        # no need to extract_body shit here
        log.warning('Geodis error 500')
        xml = get_parts(response)['start']
        obj = objectify.fromstring(xml)
        message = obj.xpath("//*[local-name() = 'message']")
        if len(message) > 0:
            message = message[0] or obj.xpath('//faultstring')[0]
            id_message = obj.xpath("//*[local-name() = 'code']")[0]
        errors = [{
            "id": id_message,
            "message": message,
        }]
        raise CarrierError(response, errors)

    def handle_200(self, response):
        """Handle response type 200."""
        parts = get_parts(response)
        xml = parts['start']

        def extract_soap(response_xml):
            obj = objectify.fromstring(response_xml)
            return obj.Body.getchildren()[0]

        payload = extract_soap(xml)
        attachement_cid = payload.codeAttachement.text[len('cid:'):]
        attachement = parts[attachement_cid]

        return {
            "body": payload,
            "parts": attachement,
            "response": response,
        }

    def handle_response(self, response):
        """Handle response of webservice."""
        if response.status_code == 500:
            return self.handle_500(response)
        elif response.status_code == 200:
            return self.handle_200(response)
        else:
            raise CarrierError(response, [{
                'id': None,
                'message': "Unexpected status code from server",
            }])



# -*- coding: utf-8 -*-
"""Mock class."""
from roulier.codec import Decoder


class TrsDecoder(Decoder):
    """Prepare output"""

    def decode(self, response, request):
        """Return {}."""

        return {
            'label': response['payload'],
            'tracking_number': request['body']['service']['shippingId'],
            'tracking_url': '',
            'annexes': [
                {
                    'name': 'meta',
                    'type': 'csv',
                    'data': response['attachment']
                }
            ]
        }



# -*- coding: utf-8 -*-
"""Geodis XML -> Python."""
from roulier.codec import Decoder


class GeodisDecoder(Decoder):
    """Geodis XML -> Python."""

    def decode(self, body, parts, output_format):
        """Geodis XML -> Python."""

        def reponse_impression_etiquette(msg, parts):
            x = {
                "tracking": {
                    "barcode": body.cabRouting,
                },
                "label": {
                    "name": "label",
                    "data": parts,
                    "type": output_format
                },
                "annexes": [
                ],
                "extra": {
                    "reseau": body.reseau,
                    "priorite": body.priorite,
                    "codeDirectionel": body.codire,
                    "cabRouting": body.cabRouting,
                    "colis": {
                        "codeUmg": body.infoColis.codumg,
                        "numero": body.infoColis.numero,
                        "cab": body.infoColis.cab,
                    }
                }
            }
            return x

        tag = body.tag
        lookup = {
            "{http://impression.service.web.etiquette.geodis.com}reponseImpressionEtiquette":
                reponse_impression_etiquette,
        }
        return lookup[tag](body, parts)



# -*- coding: utf-8 -*-
"""Implementation for trs."""
from trs_encoder import TrsEncoder
from trs_decoder import TrsDecoder
from trs_transport import TrsTransport
from roulier.carrier import Carrier


class Trs(Carrier):
    """Implementation for trs."""

    encoder = TrsEncoder()
    decoder = TrsDecoder()
    ws = TrsTransport()

    def api(self):
        """Expose how to communicate with trs."""
        return self.encoder.api()

    def get(self, data, action):
        """Run an action."""
        if action == "generateLabel":
            return self.get_label(data)
        if action == "depositSlip":
            return self.get_deposit_slip(data)
        raise Exception(
            'action %s not in %s' % (
                action, ', '.join(self.encoder.TRS_ACTIONS)))

    def get_label(self, data):
        """Generate a label."""
        request = self.encoder.encode(data, 'generateLabel')
        response = self.ws.send(request)

        return self.decoder.decode(response, request)

    def get_deposit_slip(self, data):
        """Generate a deposit slip (csv file)."""
        return self.ws.generate_deposit_slip(data)



# -*- coding: utf-8 -*-
"""Implementation for trs."""
from roulier.transport import Transport
from jinja2 import Environment, PackageLoader
from collections import OrderedDict
import unicodecsv as csv
from io import BytesIO


class TrsTransport(Transport):
    """Generate ZPL offline and csv for EDI."""

    STATUS_SUCCESS = "Success"

    def send(self, payload):
        """Call this function.

        Args:
            body: an object with a lot usefull values
        Return:
            {
                status: STATUS_SUCCES or STATUS_ERROR, (string)
                message: more info about status of result (None)
                response: (None)
                payload: usefull payload (if success) (body as string)

            }
        """
        body = payload['body']
        label = self.generate_zpl(body)
        attachment = self.map_delivery_line(body)
        return {
            "status": self.STATUS_SUCCESS,
            "message": None,
            "response": None,
            "payload": label,
            "attachment": attachment
        }

    def generate_zpl(self, body):
        env = Environment(
            loader=PackageLoader('roulier', '/carriers/trs/templates'),
            extensions=['jinja2.ext.with_'])

        template = env.get_template("trs_generateLabel.zpl")
        return template.render(
            service=body['service'],
            parcel=body['parcel'],
            from_address=body['from_address'],
            to_address=body['to_address'])

    def map_delivery_line(self, body):
        data = OrderedDict([
            (u'client', body['from_address']['company']),
            (u'siret', None),
            (u'refCommande', body['service']['reference1']),
            (u'dateEnlevement', body['service']['shippingDate']),
            (u'cr', None),
            (u'va', None),
            (u'nom', body['to_address']['name']),
            (u'adr1', body['to_address']['street1']),
            (u'adr2', body['to_address']['street2']),
            (u'cp', body['to_address']['zip']),
            (u'ville', body['to_address']['city']),
            (u'telephone', body['to_address']['phone']),
            (u'mobile', body['to_address']['phone']),
            (u'email', body['to_address']['email']),
            (u'refDest', body['service']['reference1']),
            (u'commentLiv', None),
            (u'nbConducteurs', None),
            (u'Poids', '%.f' % (body['parcel']['weight'] * 1000)),  # kg to g
            (u'nbColis', 1),  # one row per parcel
            (u'qtéFacturée1', None),
            (u'qtéFacturée2', None),
            (u'qtéFacturée3', None),
            (u'qtéFacturée4', None),
            (u'qtéFacturée5', None),
            (u'qtéFacturée6', None),
            (u'qtéFacturée7', None),
            (u'qtéFacturée8', None),
            (u'qtéFacturée9', None),
            (u'qtéFacturée10', None),
            (u'article1', None),
            (u'article2', None),
            (u'article3', None),
            (u'article4', None),
            (u'article5', None),
            (u'article6', None),
            (u'article7', None),
            (u'article8', None),
            (u'article9', None),
            (u'article10', None),
            (u'regroupement', None),
            (u'refComfour', None),
            (u'codeBarre', body['service']['shippingId']),
            (u'descColis', None),
            (u'porteur', None),
            (u'jourLivraison', None),
        ])
        return data

    def generate_deposit_slip(self, rows):
        output = BytesIO()

        # l'ordre est important
        headers = rows[0].keys()

        # l'ordre est fixé par headers
        writer = csv.DictWriter(output, headers, encoding='utf-8')
        writer.writeheader()
        writer.writerows(rows)
        return output



# -*- coding: utf-8 -*-
"""Implementation of TRS Api."""
from roulier.api import Api

TRS_LABEL_FORMAT = (
    'ZPL',
)


class TrsApi(Api):
    def _service(self):
        schema = super(TrsApi, self)._service()
        schema['labelFormat']['allowed'] = TRS_LABEL_FORMAT
        schema['labelFormat']['default'] = TRS_LABEL_FORMAT[0]
        schema['shippingId'].update({'required': True, 'empty': False})

        return schema

    def _address(self):
        schema = super(TrsApi, self)._address()
        schema['country'].update({'required': True, 'empty': False})
        schema['zip'].update({'required': True, 'empty': False})
        schema['city'].update({'required': True, 'empty': False})
        return schema

    def _from_address(self):
        schema = super(TrsApi, self)._from_address()
        schema['company'].update({'required': True, 'empty': False})
        schema['phone'].update({'required': False, 'empty': True})
        schema['street1']['required'] = False
        return schema

    def _auth(self):
        schema = super(TrsApi, self)._auth()
        schema['login'].update({'required': False, 'empty': True})
        schema['password']['required'] = False
        return schema

    def _schemas(self):
        # Santize all fields for zpl
        schemas = super(TrsApi, self)._schemas()
        for schema in schemas:
            for field in schemas[schema]:
                schemas[schema][field].update({'coerce': 'zpl'})
        return schemas



# -*- coding: utf-8 -*-
"""Transform input to trs zpl."""
from roulier.codec import Encoder
from roulier.exception import InvalidApiInput
from .trs_api import TrsApi

TRS_ACTIONS = ('generateLabel', 'generateDepositSlip')


class TrsEncoder(Encoder):
    """Transform input to trs zpl."""

    def encode(self, api_input, action):
        """Dispatcher."""
        if action == 'generateLabel':
            return self.generate_label(api_input)
        raise Exception(
            'action %s not in %s' % (action, ', '.join(TRS_ACTIONS)))

    def generate_label(self, api_input):
        """Transform input to trs zpl."""
        api = TrsApi()
        if not api.validate(api_input):
            raise InvalidApiInput(
                'Input error : %s' % api.errors(api_input))
        data = api.normalize(api_input)

        data['to_address']["dept"] = data['to_address']['zip'][0:2]

        return {'body': data, 'header': None}

    def api(self):
        api = TrsApi()
        return api.api_values()
